/** Buckets v1.00 - chaotic water wheel simulation
    Copyright (C) 1994   Tony Smith-Grieco

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as
    published by the Free Software Foundation.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    It should be in a file name COPYING.
    For information on how to reach the author, look at the file
    BUCKETS.DOC that you should also have received.
  

    I have tried to provide enough documentation so that the purpose of
    each section of code is clear - but I haven't explained what every
    single variable is for.  Hopefully that isn't necessary.

    The macro 'GR' which appears frequently throughout the program
    indicates that the program is to run in normal graphics mode.

    The macro 'DO3D' which appears occasionally is there because I
    was at one point toying with the idea of a 3-dimensional plot.
    If you compile the program with DO3D defined, you may still have
    to make a few changes to get it to work.
    
    Disclaimer:
    Don't expect this program to win marks for following the principles
    of structured programming; I was too lazy to divide the program up
    into as many little functions as humanly possible.
    
 */


#include <graphics.h>
#include <alloc.h>
#include <process.h>
#include <math.h>
#include <dos.h>
#include <stdio.h>
#include <conio.h>


/* a whole bunch of macros */


#define TRUE 1
#define FALSE 0
#define DEF_STRLEN 200

#define PROGRAM_NAME "Buckets"
#define PROGRAM_VER "1.00"

#define DEF_DATA_FILE "buckets.dat"
#define DEF_VAL -20000

#define BUCK_CLR GREEN
#define WATER_CLR CYAN
#define ATT_COLOR MAGENTA
#define CUR_ATT_CLR LIGHTGRAY
#define BORDER_CLR LIGHTGRAY

#define PAUSE_INC 10
#define SKIP_NUM 30
#define ZOOM_FACTOR 5

#define RIGHT_DIR 6
#define LEFT_DIR 4
#define UP_DIR 8
#define DOWN_DIR 2

#define PI 3.141592

#define NUM_BUCKETS 6


/* some important data types */


typedef double FP;

typedef struct bucket_info
	   {
		FP mass;
		FP angle;
	   } BUCKET;


/* millions of parameters, constants, and what not */

	   
FP pi2 = 2*PI, d_to_r = PI / 180;
FP st_ang_deg = 85, st_ang;
FP max = 2;
FP in = 1, out = .02, out_coeff = .05;
FP st_mass = 0;
FP min_ang_deg = 80, max_ang_deg = 90, min_ang, max_ang;
FP g = 9.8, r = 3, r2, gr;
FP init_ang_vel = 0;
FP fric_ang_vel = .0003;
FP fric_coeff = .3, fric_torque;
FP ang_inc_deg = 60, ang_inc;
FP pause_num = 0;
FP dt = .05;

char fric_flag = 1;
char att = 2, proj_3d = 0;
char wheelplot = 1;

int wheel_x, wheel_y, wheel_r, buck_x, buck_y;
int midx, midy, left_x, right_x, wid_x, bot_y, top_y, hgt_y;
int num_att_sections = 3;
FP dispmax;

FP max_x = 2, min_x = 0, range_x;
FP max_y = 300, min_y = -300, range_y;
FP z_dist = 70;
int x_deg_flag = 0, y_deg_flag = 1;


char proginfo[] =
    PROGRAM_NAME " v" PROGRAM_VER ", Copyright (C) 1994  Tony Smith-Grieco\n\n"
    PROGRAM_NAME " comes with ABSOLUTELY NO WARRANTY; for details,\n"
    "type 'w' at the next prompt.  This is free software, and you\n" 
    "are welcome to redistribute it under certain conditions; type\n" 
    "'c' at the next prompt for more information.\n" 
    "\nFor instructions on how to use this program, refer to the file\n" 
    "BUCKETS.DOC which should have accompanied this program.\n";

char warrantyinfo[] =
	"This program is distributed in the hope that it will be useful,\n" 
	"but WITHOUT ANY WARRANTY; without even the implied warranty of\n" 
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n" 
     "GNU General Public License, which is in a file named COPYING\n" 
	"that you should have received with this program, for more details.\n";

char licenseinfo[] =
     "If for some reason you did not receive this file, you may not\n"
     "have a complete distribution of Buckets.  Refer to BUCKETS.DOC\n" 
     "for more information.  If you don't have that file, you probably\n" 
     "should try to find another copy of Buckets.\n\n" 
	"Alternatively, to get a copy of the GNU General Public License,\n" 
	"write to the Free Software Foundation, Inc., 675 Mass Ave,\n" 
	"Cambridge, MA 02139, USA.\n";

char copyinginfo[] = 
    "This program is free software; you can redistribute it and/or modify\n" 
    "it under the terms of version 2 of the GNU General Public License as\n" 
    "published by the Free Software Foundation.  The license may be found\n" 
    "in a file named COPYING which you should have received with this\n" 
    "program.\n";


/* function prototypes */


void draw_bucket(BUCKET b[], int bnum);
char read_params(char data_file[DEF_STRLEN]);
void read_one_val(FILE *fp, FP *v);
int move_att(int dirnum);
void clear_att_window(void);
void update_axes(void);
void disp_num(int x, int y, FP num);
void set_coords(int gdriver);
void config_read_err(const char *data_file);
char *find_fname(char *full_path);

/* the program itself */


int main(int argc, char *argv[])
{
	BUCKET b[NUM_BUCKETS];
	int bnum, x, y, key, lastx = -5, lasty = -5;
	int skipcount = 0, inforate = 5, infocount = 0, magcount = 0;
	int dirnum;
	unsigned int bufsize, pause_len;
	char done = 0, dispinfo = 1, buf1[DEF_STRLEN], buf2[DEF_STRLEN];
	char found_config_file, *start_fname;
	FP force, total_force, net_torque, ang_acc, moment_inertia;
#ifdef DO3D
	FP true_torque;
#endif	
	FP ang_vel = init_ang_vel, d_theta = 0, d_angvel = 0;
	FP total_mass, scale_3d;
	FP ke, pe;						
	FP t = 0;
#ifdef GR
	int gdriver = DETECT, gmode = EGAHI;
#endif

	puts("");
	puts(proginfo);
	
/* Read parameters from data file.  If the filename is not specified on
 * the command line, first it checks the current directory, then the
 * directory where Buckets is located. */

/* First we get the full path of Buckets from the command line - if it is
 * provided in argv[0] - and determine where the path ends and the filename
 * begins. */

 	if (argv[0])
 	{
		strcpy(buf1, argv[0]);
		start_fname = find_fname(buf1);
	}
	else
	{
		buf1[0] = '\0';
		start_fname = NULL;
	}

/* Now we can look for the data. */
	
	found_config_file = FALSE;
	if (argc > 1)
		if (read_params(argv[1]))
			config_read_err(argv[1]);
		else
			found_config_file = TRUE;

	if (!found_config_file)
		if (read_params(DEF_DATA_FILE))
		{
			config_read_err(DEF_DATA_FILE);
			
			if (start_fname)
			{
				strcpy(start_fname, DEF_DATA_FILE);

				if (read_params(buf1))
					config_read_err(buf1);
				else
					found_config_file = TRUE;	
			}
		}
		else
			found_config_file = TRUE;

	if (!found_config_file)
		puts("Using default values as parameters...\n");

	done = FALSE;
	do
	{
		puts("\nPress 'q' to quit, 'w' to see warranty information,");
		puts("'c' to see copying information, or any other key to start...");
		key = tolower(getch());

		switch (key) {
			case 'q' : exit(2); break;	/* that break is probably totally
										pointless..but who cares? */
			case 'w' :
				puts("\n\n");	puts(warrantyinfo);
				puts("");  puts(licenseinfo);  break;

			case 'c' :
				puts("\n\n");  puts(copyinginfo);
				puts("");  puts(licenseinfo);  break;
				
			default  : done = TRUE; break;
		}
	} while (!done);		
	
/* Hop into graphics mode - 
 * The program looks in the directory where BUCKETS is located for the
 * BGI (graphics driver) files if they are not located in the current
 * directory. */

#ifdef GR
	*start_fname = '\0';
	initgraph(&gdriver, &gmode, buf1);

	set_coords(gdriver);
#endif

/* Initialize some variables */

	if ((pause_num >= 0) && (pause_num <= 9))
		pause_len = (9 - (int) pause_num) * PAUSE_INC;
	else
		pause_len = 0;

	if (x_deg_flag)
	{
		min_x *= d_to_r;
		max_x *= d_to_r;
	}

	if (y_deg_flag)
	{
		min_y *= d_to_r;
		max_y *= d_to_r;
	}

	midx = buck_x / 2;
	midy = buck_y / 2;
	range_y = max_y - min_y;
	range_x = max_x - min_x;
	hgt_y = bot_y - top_y + 1;
	wid_x = right_x - left_x + 1;

	r2 = r * r;  gr = g * r;
	
	fric_torque = r * fric_coeff;
	st_ang = st_ang_deg * d_to_r;
	ang_inc = ang_inc_deg * d_to_r;
	min_ang = min_ang_deg * d_to_r;
	max_ang = max_ang_deg * d_to_r;
	dispmax = max;

	for (bnum = 0; bnum < NUM_BUCKETS; bnum++)
	{
		b[bnum].mass = st_mass;
		b[bnum].angle = st_ang + (ang_inc * bnum);
	}

/* Draw border of phase space plot */

	setcolor(BORDER_CLR);
	line(left_x - 1, top_y - 1, right_x + 1, top_y - 1);
	line(left_x - 1, top_y - 1, left_x - 1, bot_y + 1);
	line(left_x - 1, bot_y + 1, right_x + 1, bot_y + 1);
	line(right_x + 1, bot_y + 1, right_x + 1, top_y - 1);
	update_axes();

/* Print some of the info at top of screen */

	sprintf(buf2,
       "dt %6.3f   Inflow %6.3f   Outflow %6.3f   Max %6.3f   Friction %6.3f",
        dt, in, out_coeff, max, fric_coeff);
	outtextxy(0, 10, buf2);
	outtextxy(0, 20, "Keys: 0-9, Alt 1-9, Arrow keys, (q)uit, (p)ause, " \
		"wcszmn");
		
/* We're ready to roll... */

	done = FALSE;

	while (1)
	{

/* Compute torques */

		net_torque = 0;
		total_mass = 0;
		total_force = 0;

#ifndef GR
		printf("##  Mass  Angle (Deg)   Force-  Torque-\n");
#endif

		for (bnum = 0; bnum < NUM_BUCKETS; bnum++)
		{
			total_force += (force = b[bnum].mass * cos(b[bnum].angle));
			total_mass += b[bnum].mass;

#ifndef GR
			printf("%2d  %4.2f  %5.3f %5.1f   %6.3f\n", bnum,
					b[bnum].mass, b[bnum].angle, b[bnum].angle / d_to_r,
					force * gr);
#else
			if (((!att) || (att == 2)) && (!skipcount) && (wheelplot))
				draw_bucket(b, bnum);
#endif
		}

/* there is a negative sign here in order to assure that a positive
   torque results in acceleration in the positive direction */
		
		net_torque = -total_force * gr;

/* Handle friction if necessary */

#ifdef DO3D
		true_torque = net_torque;
#endif
		
		if (fric_flag)
		{
			if (ang_vel > 0)
			{
				if ((net_torque < fric_torque) && (net_torque > 0))
					net_torque = 0;
				else
					net_torque -= fric_torque;
			}
			else if (ang_vel < 0)
			{
				if ((-net_torque < fric_torque) && (net_torque < 0))
					net_torque = 0;
				else
					net_torque += fric_torque;
			}
		}

/* Compute angular acceleration */

		if (total_mass > 0)
			ang_acc = net_torque / (total_mass * r2);
		else
			ang_acc = 0;

		d_angvel = ang_acc * dt;

/* Display some information if we're not in graphics mode */

#ifndef GR
		printf("   net torque  %7.3f  moment  %6.3f  ", net_torque, moment_inertia);
		printf("ang. acc  %6.3f (%6.2f)\n", ang_acc, ang_acc / d_to_r);
		printf("   dth %.5f (%6.2f)  dav %.5f (%6.2f)\n", d_theta, d_theta / d_to_r,
					d_angvel, d_angvel / d_to_r);
		printf("   angular velocity    %6.3f (%6.2f)\n\n", ang_vel, ang_vel / d_to_r);

		getchar();
#else

/* Handle user keypresses */

		if (kbhit())
		{
			key = tolower(getch());
				
			switch (key)
			{

/* Quit */
			
				case 'q' : done = 1; break;

/* Toggle the display of the wheel */

				case 'w' : wheelplot = !wheelplot; break;
				
/* Clear attractor window if params changed */

				case 'c' :
					clear_att_window();
					break;

/* Toggle information at top of screen */

				case 's' :
					dispinfo = dispinfo ^ 1;
					infocount = 0;
					break;

/* Control speed at which simulation runs */
/* (if you actually have a computer that does this too quickly) */
					
				case '1' :
				case '2' :
				case '3' :
				case '4' :
				case '5' :
				case '6' :
				case '7' :
				case '8' :
				case '9' :
				case '0' :
					if (key == '0')
						pause_len = 0;
					else
						pause_len = ('9' - key + 1) * PAUSE_INC;
					break;

/* Pause */

				case 'p' :
					while (!kbhit())
						;
					break;

				case 'z' : skipcount += SKIP_NUM; break;

/* 'Magnify' the buckets to see smaller amounts */

				case 'm' :
					dispmax /= (FP) ZOOM_FACTOR;
					magcount++;
					break;

/* 'Zoom out' (reverse of magnify) */

				case 'n' :
					dispmax *= ZOOM_FACTOR;
					magcount--;
					break;

/* Scroll attractor */

				case 0 :
					switch (key = getch())
					{
						case 77 :	dirnum = RIGHT_DIR; break;
						case 75 : dirnum = LEFT_DIR; break;
						case 80 : dirnum = DOWN_DIR; break;
						case 72 : dirnum = UP_DIR; break;

/* Alt-number keys controls how often to update the numerical display */
/* at the top of the screen */

						case 120 :
						case 121 :
						case 122 :
						case 123 :
						case 124 :
						case 125 :
						case 126 :
						case 127 :
						case 128 :
							inforate = key - 120;

						default : dirnum = 0; break;
					}

					if (dirnum)
					{

/* things get a little tricky because the most recent point to be plotted
 * is plotted a different color to indicate where in phase space the system
 * currently is; when the plot moves, this has to be taken care of. */
					
						if (lastx > 0)
						{
							putpixel(lastx, lasty, ATT_COLOR);
							lastx = -5;
						}

/* Now we can move the plot. */
						
						if (!move_att(dirnum))
							putch(7);
						else	switch (dirnum)
						{
							case RIGHT_DIR : min_x += range_x / 3; break;
							case LEFT_DIR :  min_x -= range_x / 3; break;
							case DOWN_DIR :  min_y -= range_y / 3; break;
							case UP_DIR :    min_y += range_y / 3; break;
						}

						update_axes();
					}

					break;			/* for 'case 0' */
			}
		}				/* if (kbhit()) */

		if (done)
			break;

/* Display relevant information at top of screen */

		if (!skipcount && dispinfo && !infocount)
		{
			sprintf(buf1, "%s v%s  imaginary time elapsed %8.3f  "
					"magnification factor %d",
					PROGRAM_NAME, PROGRAM_VER, t, magcount);
						
			setfillstyle(SOLID_FILL, BLACK);
 			bar(0, 0, 600, 9);

			setcolor(LIGHTGRAY);
			outtextxy(0, 0, buf1);

			infocount = inforate;
		}
		else if (!skipcount)
			infocount--;

/* Update attractor */

		if (att && (!skipcount))
		{

/* for a brief period of time I was playing around with the idea of
 * having a somewhat 3-dimensional phase space plot.  Here's the code
 * if you want to explore it some more. */
		
#ifdef DO3D
			if ((true_torque < z_dist) || !proj_3d)
			{
				if (proj_3d)
					scale_3d = z_dist / (z_dist - true_torque);
				else
					scale_3d = 1;

				x = ((total_mass - min_x - .00001) / range_x) * wid_x
					* scale_3d + left_x;
				y = bot_y - ((d_theta - min_y - .00001) / range_y)
					* hgt_y * scale_3d;
			else
				x = left_x - 10;
#else

/* Here's the code that normally is executed for a 2-d plot. */

			x = ((total_mass - min_x - .00001) / range_x) * wid_x
					+ left_x;
			y = bot_y - ((ang_vel - min_y - .00001) / range_y)
					* hgt_y;
			
#endif

/* The most recent point is plotted a different color, in order to 
 * distinguish the current state of the system, so the 'old' most
 * recent point has to be replotted a normal color. */
 
			if (lastx > 0)
				putpixel(lastx, lasty, ATT_COLOR);

			if ((x >= left_x) && (x <= right_x) && (y >= top_y) && (y <= bot_y))
			{

/* this commented-out section makes the program draw lines connecting
 * successive points which are far apart; makes the graph a little
 * prettier and smoother. */
			
/*				if ((lastx > 0) && (abs(lastx-x) < 40) && (abs(lasty-y) < 40))
				{
					setcolor(ATT_COLOR);
					line(lastx, lasty, x, y);
				}
				else
*/

				putpixel(x, y, CUR_ATT_CLR);

				lastx = x;
				lasty = y;
			}
			else
			{
				lastx = -5;
				lasty = -5;
			}
		}

/* A little pause if the user so desires */
		
		delay(pause_len);

#endif  					/* matches #ifndef GR way, way back there */

/* Update angular velocity */

		ang_vel += d_angvel;
		d_theta = ang_vel * dt;

/* Update bucket positions, handle water flowing in/out */

		for (bnum = 0; bnum < NUM_BUCKETS; bnum++)
		{
			b[bnum].angle += d_theta;

			while (b[bnum].angle < 0)
				b[bnum].angle += pi2;
			while (b[bnum].angle >= pi2)
				b[bnum].angle -= pi2;

			if ((b[bnum].angle >= min_ang) && (b[bnum].angle <= max_ang))
				b[bnum].mass += in * dt;
			if (b[bnum].mass > max)
				b[bnum].mass = max;

			b[bnum].mass -= out_coeff * b[bnum].mass * dt;
		}

/* Miscellaneous counters */
		
		t += dt;
		if (skipcount)
			skipcount--;
	}

#ifdef GR
	closegraph();
#endif

	return 0;
}

/** draw_bucket()
 *
 * Draws one of the buckets
 */
 
void draw_bucket(BUCKET b[], int bnum)
{
#ifdef GR
	int x, y, numb;
	static int oldx[NUM_BUCKETS], oldy[NUM_BUCKETS];
	static int first = 1;
	int depth;
	BUCKET binf;

	binf = b[bnum];
	x = wheel_x + wheel_r * cos(binf.angle);
	y = wheel_y - wheel_r * sin(binf.angle);

	if ((!first) && (bnum == 0))
	{
		for (numb = 0; numb < NUM_BUCKETS; numb++)
		{
			setfillstyle(SOLID_FILL, BLACK);
			setcolor(BLACK);
			bar(oldx[numb], oldy[numb] - buck_y, oldx[numb] + buck_x, oldy[numb]);
			line(wheel_x, wheel_y, oldx[numb] + midx, oldy[numb] - midy);
		}
	}

	setcolor(BUCK_CLR);
	line(wheel_x, wheel_y, x, y);

	x -= midx;
	y += midy;

	line(x, y, x, y - buck_y);
	line(x, y, x + buck_x, y);
	line(x + buck_x, y, x + buck_x, y - buck_y);

	if (binf.mass >= dispmax)
		depth = buck_y;
	else
	{
		depth = binf.mass * buck_y / dispmax;
		depth--;
	}
	
	if (depth > 0)
	{
		setfillstyle(SOLID_FILL, WATER_CLR);
		setcolor(WATER_CLR);
		bar(x + 1, y - depth, x + buck_x - 1, y - 1);
	}

	oldx[bnum] = x;
	oldy[bnum] = y;
	first = 0;
#endif
}

/** read_one_val()
 *
 * Reads one particular value from a configuration file, using the
 * current value of the variable v points to as a default.
 */
 
void read_one_val(FILE *fp, FP *v)
{
	char val_str[80] = "# ";
	FP temp_v;

	while (!feof(fp)  &&  (*val_str == '#'))
		fgets(val_str, 80, fp);

	temp_v = atof(val_str);
	if (temp_v != DEF_VAL)
		*v = temp_v;
}

/** read_params()
 *
 * Reads an entire configuration file, or belches if it can't.
 */
 
char read_params(char data_file[DEF_STRLEN])
{
	FILE *fp;
	char problem = FALSE;
	
	fp = fopen(data_file, "rt");
	if (!fp)
	{
		return 1;
	}

	printf("\n\nReading parameters from file %s... ", data_file);

	read_one_val(fp, &in);
	read_one_val(fp, &out_coeff);
	read_one_val(fp, &max);
	read_one_val(fp, &st_ang_deg);
	read_one_val(fp, &st_mass);
	read_one_val(fp, &ang_inc_deg);
	read_one_val(fp, &min_ang_deg);
	read_one_val(fp, &max_ang_deg);
	read_one_val(fp, &fric_coeff);
	read_one_val(fp, &pause_num);
	read_one_val(fp, &dt);
	read_one_val(fp, &min_x);
	read_one_val(fp, &max_x);
	read_one_val(fp, &min_y);
	read_one_val(fp, &max_y);

	if (min_x == max_x)
	{
		puts("\nThe minimum and maximum values for the horizontal");
		printf("axis can't be the same.  They are both %10.5f.\n", min_x);
		puts("You might want to change this...");

		problem = TRUE;
	}

	if (min_y == max_y)
	{
		puts("\nThe minimum and maximum values for the vertical");
		printf("axis can't be the same.  They are both %10.5f.\n", min_y);
		puts("You might want to change this...");

		problem = TRUE;
	}

	if (problem)
	{
		exit(1);
	}
	
#ifdef DISP_PARAMS
	printf("\nin %8.3f out %8.3f sad %f sm %f aid %f\n", in, out_coeff,
			st_ang_deg, st_mass, ang_inc_deg);
	printf("bucket size %f mid %f mad %f fc %f pn %f dt %f", max,
			min_ang_deg, max_ang_deg, fric_coeff, pause_num, dt);

	puts("\n\nPress any key to start the simulation...");
	getch();
#endif

/* To finish off the "Reading configuration..." line earlier */

	puts("Finished.\n");
	
	return 0;
}

/** move_att()
 *
 * Does the arduous task of shifting what is currently on the graph
 * in the appropriate direction.
 *
 * This code looks so ugly because this is a real pain to do.  I don't
 * have enough memory to grab the whole plot in one chunk (at least in EGA
 * and VGA it's a problem) so I have to divide the chunk into three slices
 * and move them one at a time, making sure to move them in the right order
 * with no overlap.... you get the idea.
 */
 
int move_att(int dirnum)
{
	char *b, buf[80], done;
	int mleft_x, mright_x, mbot_y, mtop_y, mwid, msection_wid;
	int new_left_x, new_top_y, new_right_x, new_bot_y;
	int cmtop_y, cmbot_y, cnew_top_y;
	unsigned int bufsize, freemem;
	unsigned long farfreemem;

	mleft_x = left_x;
	mright_x = right_x;
	mtop_y = top_y;
	mbot_y = bot_y;

	new_left_x = left_x;
	new_right_x = right_x;
	new_top_y = top_y;
	new_bot_y = bot_y;

/* The variables which start with 'm' relate to the area that is
 * currently on the screen that will be moved.  Those that start with
 * 'new' indicate where that chunk will be moved to. */
	
	switch (dirnum)
	{
		case RIGHT_DIR :
			mleft_x = wid_x / 3 + left_x;
			new_right_x = right_x - (mleft_x - left_x);
			break;

		case LEFT_DIR :
			mright_x = right_x - wid_x / 3;
			new_left_x = right_x - (mright_x - mleft_x);
			break;

		case UP_DIR :
			mbot_y = bot_y - hgt_y / 3;
			new_top_y = bot_y - (mbot_y - mtop_y);
			break;

		case DOWN_DIR :
			mtop_y = hgt_y / 3 + top_y;
			new_bot_y = bot_y - (mtop_y - top_y);
			break;
	}

	mwid = mbot_y - mtop_y + 1;
	msection_wid = mwid / num_att_sections + 1;
	bufsize = imagesize(mleft_x, mtop_y, mright_x, mtop_y + msection_wid - 1);
	b = (char *) malloc(bufsize);

	if (!b)
	{
		putch(7);
		return(0);
	}

/* I wish you luck trying to understand this mess.  Basically this next
 * if block just 'undoes' what the do loop will do so that when the do
 * loop actually executes for the first time, it starts in the right
 * place. */
	
	if (dirnum == UP_DIR)
	{
		cmtop_y = mbot_y + 1;
		cmbot_y = mbot_y + msection_wid;
		cnew_top_y = bot_y + 1;
	}
	else
	{
		cmtop_y = mtop_y - msection_wid;
		cmbot_y = mtop_y - 1;
		cnew_top_y = new_top_y - msection_wid;
	}

/* Now we actually move the schlock, section by section. */

	done = FALSE;	
	do
	{
		if (dirnum == UP_DIR)
		{
			cmbot_y -= msection_wid;
			cmtop_y -= msection_wid;
			cnew_top_y -= msection_wid;

			if (cmtop_y < mtop_y)
			{
				cmtop_y = mtop_y;
				done = TRUE;
			}
		}
		else
		{
			cmbot_y += msection_wid;
			cmtop_y += msection_wid;
			cnew_top_y += msection_wid;

			if (cmbot_y > mbot_y)
			{
				cmbot_y = mbot_y;
				done = TRUE;
			}
		}

		getimage(mleft_x, cmtop_y, mright_x, cmbot_y, b);
		putimage(new_left_x, cnew_top_y, b, COPY_PUT);
	} while ( !done );

/* Blank out the correct part of the plot. */

	setfillstyle(SOLID_FILL, BLACK);
	switch (dirnum)
	{
		case RIGHT_DIR :
			bar(new_right_x + 1, mtop_y, right_x, mbot_y); break;

		case LEFT_DIR :
			bar(left_x, mtop_y, new_left_x - 1, mbot_y); break;

		case UP_DIR :
			bar(mleft_x, top_y, mright_x, new_top_y - 1); break;

		case DOWN_DIR :
			bar(mleft_x, new_bot_y + 1, mright_x, mbot_y); break;
	}

	free(b);

	return(1);
}

/** clear_att_window()
 *
 * Blanks out the entire phase space plot.
 */
 
void clear_att_window(void)
{
	setfillstyle(SOLID_FILL, BLACK);
	bar(left_x, top_y, right_x, bot_y);
}

/** update_axes()
 *
 * Redisplays the numerals along the axes which indicate the current
 * bounds of the plot.
 */
 
void update_axes(void)
{
	FP x_divnum = 1, y_divnum = 1;

	if (x_deg_flag)
		x_divnum = d_to_r;

	if (y_deg_flag)
		y_divnum = d_to_r;

	setfillstyle(SOLID_FILL, BLACK);
	bar(left_x, bot_y + 2, left_x + 80, bot_y + 12);
	bar(right_x - 80, bot_y + 2, right_x, bot_y + 12);
	bar(right_x + 5, top_y, right_x + 80, top_y + 10);
	bar(right_x + 5, bot_y - 10, right_x + 80, bot_y);

	setcolor(BORDER_CLR);

	disp_num(left_x, bot_y + 4, min_x / x_divnum);
	disp_num(right_x + 5, top_y, (max_y = min_y + range_y) / y_divnum);
	disp_num(right_x + 5, bot_y - 10, min_y / y_divnum);

	settextjustify(RIGHT_TEXT, TOP_TEXT);
	disp_num(right_x, bot_y + 4, (max_x = min_x + range_x) / x_divnum);
	settextjustify(LEFT_TEXT, TOP_TEXT);
}

/** disp_num()
 *
 * Stupid little thing to display a single number on the screen.
 */
 
void disp_num(int x, int y, FP num)
{
	char b[50];

	sprintf(b, "%7.3f", num);
	outtextxy(x, y, b);
}

/** set_coords()
 *
 * Based on what graphics mode we're using, determine screen
 * coordinates for various parts of the display.
 */
 
void set_coords(int gdriver)
{
	switch (gdriver)
	{
		case CGA:
			wheel_x = 530;  wheel_y = 100;  wheel_r = 45;
			buck_x = 15;  buck_y = 20;
			left_x = 10;  right_x = 460;
			bot_y = 190;  top_y = 30;
			break;

		case VGA:
			wheel_x = 550;  wheel_y = 240;  wheel_r = 50;
			buck_x = 15;  buck_y = 35;
			left_x = 20;  right_x = 460;
			bot_y = 450;  top_y = 50;
			break;

		case EGA:
			wheel_x = 550;  wheel_y = 175;  wheel_r = 40;
			buck_x = 15;  buck_y = 35;
			left_x = 20;  right_x = 460;
			bot_y = 330;  top_y = 30;
			break;
	}
}

/** config_read_err()
 *
 * If you can't figure out what this does, you shouldn't be trying
 * to mess with this source code.
 */
 
void config_read_err(const char *data_file)
{
	printf("I can't read the configuration file %s.\n", data_file);
}

/** find_fname()
 *
 * Given a filename with a complete or partially complete path, this
 * returns a pointer indicating where the filename itself actually
 * begins.
 *
 * For example, it locates the 'FERENGI.FUN' part of this path:
 *                \STAR\TREK\RACES\FERENGI.FUN
 */
 
char *find_fname(char *full_path)
{
	char *p, *file_pos;

	p = full_path;
	file_pos = full_path;

	while (*p)
		if ( *p == ':' || *p == '\\' )
			file_pos = ++p;
		else
			p++;

	return file_pos;
}

